# autogen.sh configuration for GnuPG                           -*- sh -*-

case "$myhost:$myhostsub" in
  w32:ce)
    extraoptions=
    ;;
  w32:)
    extraoptions=
    ;;
esac

case "$myhost" in
  w32)
    configure_opts=""
    ;;

  amd64)
    configure_opts=""
    ;;
esac


extra_aclocal_flags="-I gl/m4"

final_info="./configure --enable-maintainer-mode && make"
