/* resource.h
 *       Copyright 2004 g10 Code GmbH
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * SPDX-License-Identifier: FSFULLR
 */

#define IDD_PINENT                      101
#define IDC_PINENT_TEXT                 292
#define IDC_PINENT_DESC                 1000
#define IDC_PINENT_ERR                  1001
#define IDC_PINENT_PROMPT               1002
#define IDC_PINENT_ICON                 1003

#define IDB_ICON_32                     3101
#define IDB_ICON_48                     3102
#define IDB_ICON_64                     3103
#define IDB_ICON_96                     3104
#define IDB_ICON_128                    3105
