; libtnbtls.def - List of symbols to export.
; Copyright (C) 2014 g10 Code GmbH
;
; This file is part of NTBTLS
;
; NTBTLS is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 3 of the License, or
; (at your option) any later version.
;
; NTBTLS is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, see <http://www.gnu.org/licenses/>.
;
; NOTE: When adding new functions, please make sure to add them also
;       to visibility.{c,h} and libntbtls.vers.


EXPORTS
    ntbtls_check_version                  @1
    ntbtls_set_debug                      @2
    ntbtls_set_log_handler                @3

    ntbtls_new                            @4
    _ntbtls_check_context                 @5
    ntbtls_release                        @6

    ntbtls_set_transport                  @7
    ntbtls_get_stream                     @8
    ntbtls_set_hostname                   @9
    ntbtls_get_hostname                   @10
    ntbtls_set_verify_cb                  @11

    ntbtls_handshake                      @12

    ntbtls_x509_get_peer_cert             @13

    ntbtls_get_last_alert                 @14

; END
