/* dialog.rc - Resource definitions                             -*- c -*-
 * Copyright (C) 2004, 2010 g10 Code GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <config.h>
#include <winuser.h>
#include <windows.h>
#include "resource.h"

/*
 * Main dialog
 */

IDB_ICON_32    BITMAP   DISCARDABLE   "logo-32.bmp"
IDB_ICON_48    BITMAP   DISCARDABLE   "logo-48.bmp"
IDB_ICON_64    BITMAP   DISCARDABLE   "logo-64.bmp"
IDB_ICON_96    BITMAP   DISCARDABLE   "logo-96.bmp"
IDB_ICON_128   BITMAP   DISCARDABLE   "logo-128.bmp"

IDD_PINENT DIALOG DISCARDABLE  0, 0, 230, 125
STYLE DS_MODALFRAME | DS_SYSMODAL | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Pinentry"
FONT 10, "MS Sans Serif"
BEGIN
    CONTROL         "", IDC_PINENT_ICON,
                    "Static", SS_BITMAP|SS_CENTERIMAGE,
                                           5,  5,  32, 32
    LTEXT           "", IDC_PINENT_DESC,  45,  5, 180, 65
    RTEXT           "", IDC_PINENT_PROMPT, 5, 75,  60, 12
    EDITTEXT        IDC_PINENT_TEXT,      70, 75, 155, 12, ES_PASSWORD | ES_AUTOHSCROLL
    CTEXT           "", IDC_PINENT_ERR,    5, 90, 220, 12
    DEFPUSHBUTTON   "O&K", IDOK,          50, 105, 85, 14
    PUSHBUTTON      "&Cancel", IDCANCEL, 140, 105, 85, 14
END
